@@ -1,3 +1,6 @@
+// Copyright (c) HashiCorp, Inc.
+// SPDX-License-Identifier: BUSL-1.1
+
 package bindingruleupdate
 
 import (
@@ -37,7 +40,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	acl {
 		enabled = true
 		tokens {
-			master = "root"
+			initial_management = "root"
 		}
 	}`)
 
@@ -124,15 +127,24 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 		require.Contains(t, ui.ErrorWriter.String(), "Binding rule not found with ID")
 	})
 
-	createRule := func(t *testing.T) string {
+	createRule := func(t *testing.T, isTemplatedPolicy bool) string {
+		bindingRule := &api.ACLBindingRule{
+			AuthMethod:  "test",
+			Description: "test rule",
+			BindType:    api.BindingRuleBindTypeService,
+			BindName:    "test-${serviceaccount.name}",
+			Selector:    "serviceaccount.namespace==default",
+		}
+		if isTemplatedPolicy {
+			bindingRule.BindType = api.BindingRuleBindTypeTemplatedPolicy
+			bindingRule.BindName = api.ACLTemplatedPolicyServiceName
+			bindingRule.BindVars = &api.ACLTemplatedPolicyVariables{
+				Name: "test-${serviceaccount.name}",
+			}
+		}
+
 		rule, _, err := client.ACL().BindingRuleCreate(
-			&api.ACLBindingRule{
-				AuthMethod:  "test",
-				Description: "test rule",
-				BindType:    api.BindingRuleBindTypeService,
-				BindName:    "test-${serviceaccount.name}",
-				Selector:    "serviceaccount.namespace==default",
-			},
+			bindingRule,
 			&api.WriteOptions{Token: "root"},
 		)
 		require.NoError(t, err)
@@ -158,7 +170,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 				m[c] = struct{}{}
 			}
 
-			_ = createRule(t)
+			_ = createRule(t, false)
 		}
 	}
 
@@ -180,7 +192,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("must use roughly valid selector", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		args := []string{
 			"-http-addr=" + a.HTTPAddr(),
@@ -198,7 +210,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -230,10 +242,112 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
 	})
 
+	t.Run("update all fields with policy", func(t *testing.T) {
+		id := createRule(t, false)
+
+		ui := cli.NewMockUi()
+		cmd := New(ui)
+
+		args := []string{
+			"-http-addr=" + a.HTTPAddr(),
+			"-token=root",
+			"-id", id,
+			"-description=test rule edited",
+			"-bind-type", "policy",
+			"-bind-name=policy-updated",
+			"-selector=serviceaccount.namespace==alt and serviceaccount.name==demo",
+		}
+
+		code := cmd.Run(args)
+		require.Equal(t, code, 0, "err: %s", ui.ErrorWriter.String())
+		require.Empty(t, ui.ErrorWriter.String())
+
+		rule, _, err := client.ACL().BindingRuleRead(
+			id,
+			&api.QueryOptions{Token: "root"},
+		)
+		require.NoError(t, err)
+		require.NotNil(t, rule)
+
+		require.Equal(t, "test rule edited", rule.Description)
+		require.Equal(t, "policy-updated", rule.BindName)
+		require.Equal(t, api.BindingRuleBindTypePolicy, rule.BindType)
+		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
+	})
+
+	t.Run("update all fields with templated policy", func(t *testing.T) {
+		id := createRule(t, false)
+
+		ui := cli.NewMockUi()
+		cmd := New(ui)
+
+		args := []string{
+			"-http-addr=" + a.HTTPAddr(),
+			"-token=root",
+			"-id", id,
+			"-description=test rule edited",
+			"-bind-type", "templated-policy",
+			"-bind-name=builtin/service",
+			"-bind-vars", "name=api",
+			"-selector=serviceaccount.namespace==alt and serviceaccount.name==demo",
+		}
+
+		code := cmd.Run(args)
+		require.Equal(t, code, 0, "err: %s", ui.ErrorWriter.String())
+		require.Empty(t, ui.ErrorWriter.String())
+
+		rule, _, err := client.ACL().BindingRuleRead(
+			id,
+			&api.QueryOptions{Token: "root"},
+		)
+		require.NoError(t, err)
+		require.NotNil(t, rule)
+
+		require.Equal(t, "test rule edited", rule.Description)
+		require.Equal(t, api.ACLTemplatedPolicyServiceName, rule.BindName)
+		require.Equal(t, api.BindingRuleBindTypeTemplatedPolicy, rule.BindType)
+		require.Equal(t, &api.ACLTemplatedPolicyVariables{Name: "api"}, rule.BindVars)
+		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
+	})
+
+	t.Run("update bind type to something other than templated-policy unsets bindvars", func(t *testing.T) {
+		id := createRule(t, true)
+
+		ui := cli.NewMockUi()
+		cmd := New(ui)
+
+		args := []string{
+			"-http-addr=" + a.HTTPAddr(),
+			"-token=root",
+			"-id", id,
+			"-description=test rule edited",
+			"-bind-type", "role",
+			"-bind-name=role-updated",
+			"-selector=serviceaccount.namespace==alt and serviceaccount.name==demo",
+		}
+
+		code := cmd.Run(args)
+		require.Equal(t, code, 0, "err: %s", ui.ErrorWriter.String())
+		require.Empty(t, ui.ErrorWriter.String())
+
+		rule, _, err := client.ACL().BindingRuleRead(
+			id,
+			&api.QueryOptions{Token: "root"},
+		)
+		require.NoError(t, err)
+		require.NotNil(t, rule)
+
+		require.Equal(t, "test rule edited", rule.Description)
+		require.Equal(t, "role-updated", rule.BindName)
+		require.Equal(t, api.BindingRuleBindTypeRole, rule.BindType)
+		require.Empty(t, rule.BindVars)
+		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
+	})
+
 	t.Run("update all fields - partial", func(t *testing.T) {
 		deleteRules(t) // reset since we created a bunch that might be dupes
 
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -266,7 +380,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields but description", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -298,7 +412,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields but bind name", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -330,7 +444,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields but must exist", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -362,7 +476,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields but selector", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -394,7 +508,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields clear selector", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -427,7 +541,7 @@ func TestBindingRuleUpdateCommand(t *testing.T) {
 	})
 
 	t.Run("update all fields json formatted", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -478,7 +592,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 	acl {
 		enabled = true
 		tokens {
-			master = "root"
+			initial_management = "root"
 		}
 	}`)
 
@@ -568,15 +682,25 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 		require.Contains(t, ui.ErrorWriter.String(), "Binding rule not found with ID")
 	})
 
-	createRule := func(t *testing.T) string {
+	createRule := func(t *testing.T, isTemplatedPolicy bool) string {
+		bindingRule := &api.ACLBindingRule{
+			AuthMethod:  "test",
+			Description: "test rule",
+			BindType:    api.BindingRuleBindTypeRole,
+			BindName:    "test-${serviceaccount.name}",
+			Selector:    "serviceaccount.namespace==default",
+		}
+
+		if isTemplatedPolicy {
+			bindingRule.BindType = api.BindingRuleBindTypeTemplatedPolicy
+			bindingRule.BindName = "builtin/service"
+			bindingRule.BindVars = &api.ACLTemplatedPolicyVariables{
+				Name: "test-${serviceaccount.name}",
+			}
+		}
+
 		rule, _, err := client.ACL().BindingRuleCreate(
-			&api.ACLBindingRule{
-				AuthMethod:  "test",
-				Description: "test rule",
-				BindType:    api.BindingRuleBindTypeRole,
-				BindName:    "test-${serviceaccount.name}",
-				Selector:    "serviceaccount.namespace==default",
-			},
+			bindingRule,
 			&api.WriteOptions{Token: "root"},
 		)
 		require.NoError(t, err)
@@ -602,7 +726,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 				m[c] = struct{}{}
 			}
 
-			_ = createRule(t)
+			_ = createRule(t, false)
 		}
 	}
 
@@ -625,7 +749,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 	})
 
 	t.Run("must use roughly valid selector", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -647,7 +771,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 	})
 
 	t.Run("update all fields", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -680,10 +804,81 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
 	})
 
+	t.Run("update all fields after initial binding rule is templated-policy", func(t *testing.T) {
+		id := createRule(t, true)
+
+		ui := cli.NewMockUi()
+		cmd := New(ui)
+
+		args := []string{
+			"-http-addr=" + a.HTTPAddr(),
+			"-token=root",
+			"-no-merge",
+			"-id", id,
+			"-description=test rule edited",
+			"-bind-type", "service",
+			"-bind-name=role-updated",
+			"-selector=serviceaccount.namespace==alt and serviceaccount.name==demo",
+		}
+
+		code := cmd.Run(args)
+		require.Equal(t, code, 0, "err: %s", ui.ErrorWriter.String())
+		require.Empty(t, ui.ErrorWriter.String())
+
+		rule, _, err := client.ACL().BindingRuleRead(
+			id,
+			&api.QueryOptions{Token: "root"},
+		)
+		require.NoError(t, err)
+		require.NotNil(t, rule)
+
+		require.Equal(t, "test rule edited", rule.Description)
+		require.Equal(t, api.BindingRuleBindTypeService, rule.BindType)
+		require.Empty(t, rule.BindVars)
+		require.Equal(t, "role-updated", rule.BindName)
+		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
+	})
+
+	t.Run("update all fields with templated-policy bind type", func(t *testing.T) {
+		id := createRule(t, false)
+
+		ui := cli.NewMockUi()
+		cmd := New(ui)
+
+		args := []string{
+			"-http-addr=" + a.HTTPAddr(),
+			"-token=root",
+			"-no-merge",
+			"-id", id,
+			"-description=test rule edited",
+			"-bind-type=templated-policy",
+			"-bind-name=builtin/service",
+			"-bind-vars", "name=api",
+			"-selector=serviceaccount.namespace==alt and serviceaccount.name==demo",
+		}
+
+		code := cmd.Run(args)
+		require.Equal(t, code, 0, "err: %s", ui.ErrorWriter.String())
+		require.Empty(t, ui.ErrorWriter.String())
+
+		rule, _, err := client.ACL().BindingRuleRead(
+			id,
+			&api.QueryOptions{Token: "root"},
+		)
+		require.NoError(t, err)
+		require.NotNil(t, rule)
+
+		require.Equal(t, "test rule edited", rule.Description)
+		require.Equal(t, api.BindingRuleBindTypeTemplatedPolicy, rule.BindType)
+		require.Equal(t, api.ACLTemplatedPolicyServiceName, rule.BindName)
+		require.Equal(t, &api.ACLTemplatedPolicyVariables{Name: "api"}, rule.BindVars)
+		require.Equal(t, "serviceaccount.namespace==alt and serviceaccount.name==demo", rule.Selector)
+	})
+
 	t.Run("update all fields - partial", func(t *testing.T) {
 		deleteRules(t) // reset since we created a bunch that might be dupes
 
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -717,7 +912,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 	})
 
 	t.Run("update all fields but description", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -750,7 +945,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 	})
 
 	t.Run("missing bind name", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
@@ -770,7 +965,7 @@ func TestBindingRuleUpdateCommand_noMerge(t *testing.T) {
 	})
 
 	t.Run("update all fields but selector", func(t *testing.T) {
-		id := createRule(t)
+		id := createRule(t, false)
 
 		ui := cli.NewMockUi()
 		cmd := New(ui)
